@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-#  Copyright 2018-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
+#  Copyright 2018-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
 #  This file is part of CGRtools.
 #
 #  CGRtools is free software; you can redistribute it and/or modify
@@ -16,51 +16,139 @@
 #  You should have received a copy of the GNU Lesser General Public License
 #  along with this program; if not, see <https://www.gnu.org/licenses/>.
 #
+from CachedMethods import cached_property
 from collections import defaultdict, deque
-from typing import List, Optional, Tuple
+from typing import List, Optional, Tuple, TYPE_CHECKING, Union
 from .._functions import lazy_product
+from ..containers import query  # cyclic imports resolve
 from ..exceptions import InvalidAromaticRing
 
 
+if TYPE_CHECKING:
+    from CGRtools import MoleculeContainer
+
+
 class Aromatize:
     __slots__ = ()
 
-    def thiele(self) -> bool:
+    def thiele(self: 'MoleculeContainer', *, fix_tautomers=True) -> bool:
         """
         Convert structure to aromatic form (Huckel rule ignored). Return True if found any kekule ring.
-        Also marks atoms as aromatic. For pyroles, furans, thiophene, phospholes bonds kepts in Kekule form.
+        Also marks atoms as aromatic.
+
+        :param fix_tautomers: try to fix condensed rings with pyroles.
+            N1C=CC2=NC=CC2=C1>>N1C=CC2=CN=CC=C12
         """
         atoms = self._atoms
         bonds = self._bonds
         sh = self._hybridizations
+        charges = self._charges
+        hydrogens = self._hydrogens
 
         rings = defaultdict(set)  # aromatic? skeleton. include quinones
+        tetracycles = []
+        pyroles = set()
+        acceptors = set()
+        donors = []
+        freaks = []
         for ring in self.sssr:
-            if all(sh[n] == 2 and atoms[n].atomic_number in (5, 6, 7, 15) for n in ring):
-                n, *_, m = ring
-                rings[n].add(m)
-                rings[m].add(n)
-                for n, m in zip(ring, ring[1:]):
+            lr = len(ring)
+            if not 3 < lr < 8:  # skip 3-membered and big rings
+                continue
+            sp2 = sum(sh[n] == 2 and atoms[n].atomic_number in (5, 6, 7, 15) for n in ring)
+            if sp2 == lr:  # benzene like
+                if lr == 4:  # two bonds condensed aromatic rings
+                    tetracycles.append(ring)
+                else:
+                    if fix_tautomers and lr % 2:  # find potential pyroles
+                        n = next((n for n in ring if atoms[n].atomic_number == 7 and not charges[n]), None)
+                        if n:
+                            acceptors.add(n)
+                    n, *_, m = ring
                     rings[n].add(m)
                     rings[m].add(n)
+                    for n, m in zip(ring, ring[1:]):
+                        rings[n].add(m)
+                        rings[m].add(n)
+            elif 4 < lr == sp2 + 1:  # pyroles, furanes, etc
+                try:
+                    n = next(n for n in ring if sh[n] == 1)
+                except StopIteration:  # exotic, just skip
+                    continue
+                if atoms[n].atomic_number in (5, 7, 8, 15, 16, 34) and not charges[n]:
+                    if lr == 7 and atoms[n].atomic_number != 5:  # skip electron-rich 7-membered rings
+                        continue
+                    if fix_tautomers and lr == 6 and atoms[n].atomic_number == 7 and len(bonds[n]) == 2:
+                        donors.append(n)
+                    pyroles.add(n)
+                    n, *_, m = ring
+                    rings[n].add(m)
+                    rings[m].add(n)
+                    for n, m in zip(ring, ring[1:]):
+                        rings[n].add(m)
+                        rings[m].add(n)
+            # like N1C=Cn2cccc12
+            elif lr == 5 and sum(atoms[x].atomic_number == 7 and not charges[x] for x in ring) > 1:
+                freaks.append(ring)
         if not rings:
             return False
+        double_bonded = {n for n in rings if any(m not in rings and b.order == 2 for m, b in bonds[n].items())}
+
+        # fix_tautomers
+        if fix_tautomers and acceptors and donors:
+            for start in donors:
+                stack = [(start, n, 0, 2) for n in rings[start] if n not in double_bonded]
+                path = []
+                seen = {start}
+                while stack:
+                    last, current, depth, order = stack.pop()
+                    if len(path) > depth:
+                        seen.difference_update(x for _, x, _ in path[depth:])
+                        path = path[:depth]
+                    path.append((last, current, order))
+                    if current in acceptors:  # we found
+                        if order == 1:
+                            acceptors.discard(current)
+                            pyroles.discard(start)
+                            pyroles.add(current)
+                            hydrogens[current] = 1
+                            hydrogens[start] = 0
+                            break
+                        else:
+                            continue
+
+                    depth += 1
+                    seen.add(current)
+                    new_order = 1 if order == 2 else 2
+                    stack.extend((current, n, depth, new_order) for n in rings[current] if
+                                 n not in seen and n not in double_bonded and bonds[current][n].order == order)
+                for n, m, o in path:
+                    bonds[n][m]._Bond__order = o
+                if not acceptors:
+                    break
 
-        double_bonded = [n for n in rings if any(m not in rings and b.order == 2 for m, b in bonds[n].items())]
         if double_bonded:  # delete quinones
             for n in double_bonded:
                 for m in rings.pop(n):
                     rings[m].discard(n)
+
+            for n in [n for n, ms in rings.items() if not ms]:  # imide leads to isolated atoms
+                del rings[n]
+            if not rings:
+                return False
             while True:
                 try:
                     n = next(n for n, ms in rings.items() if len(ms) == 1)
                 except StopIteration:
                     break
                 m = rings.pop(n).pop()
-                pm = rings.pop(m)
-                pm.discard(n)
-                for x in pm:
-                    rings[x].discard(m)
+                if n in pyroles:
+                    rings[m].discard(n)
+                else:
+                    pm = rings.pop(m)
+                    pm.discard(n)
+                    for x in pm:
+                        rings[x].discard(m)
         if not rings:
             return False
 
@@ -72,18 +160,40 @@ class Aromatize:
         seen = set()
         for ring in rings:
             seen.update(ring)
+        for n in seen:
+            sh[n] = 4
+
+        # reset bonds to single
+        for ring in tetracycles:
+            if seen.issuperset(ring):
+                n, *_, m = ring
+                bonds[n][m]._Bond__order = 1
+                for n, m in zip(ring, ring[1:]):
+                    bonds[n][m]._Bond__order = 1
+
+        for ring in rings:
             n, *_, m = ring
             bonds[n][m]._Bond__order = 4
             for n, m in zip(ring, ring[1:]):
                 bonds[n][m]._Bond__order = 4
-        for n in seen:
-            sh[n] = 4
 
         self.flush_cache()
+        for ring in freaks:  # aromatize rule based
+            rs = set(ring)
+            for q in self.__freaks:
+                components, closures = q._compiled_query
+                if any(q._get_mapping(components[0], closures, atoms, bonds, rs, self.atoms_order)):
+                    n, *_, m = ring
+                    bonds[n][m]._Bond__order = 4
+                    for n, m in zip(ring, ring[1:]):
+                        bonds[n][m]._Bond__order = 4
+                    for n in ring:
+                        sh[n] = 4
+
         self._fix_stereo()  # check if any stereo centers vanished.
         return True
 
-    def kekule(self) -> bool:
+    def kekule(self: Union['Aromatize', 'MoleculeContainer']) -> bool:
         """
         Convert structure to kekule form. Return True if found any aromatic ring. Set implicit hydrogen count and
         hybridization marks on atoms.
@@ -98,7 +208,7 @@ class Aromatize:
             return True
         return False
 
-    def enumerate_kekule(self):
+    def enumerate_kekule(self: Union['Aromatize', 'MoleculeContainer']):
         """
         Enumerate all possible kekule forms of molecule.
         """
@@ -122,13 +232,41 @@ class Aromatize:
             return False
         return True
 
-    def __prepare_rings(self):
+    def __fix_oxides(self: 'MoleculeContainer'):
+        atoms = self._atoms
+        bonds = self._bonds
+        atoms_order = self.atoms_order
+        connected_components = [set(x) for x in self.connected_components]
+
+        seen = set()
+        for q, af, bf in self.__oxyde_rules:
+            components, closures = q._compiled_query
+            for candidate in connected_components:
+                for mapping in q._get_mapping(components[0], closures, atoms, bonds, candidate - seen, atoms_order):
+                    match = set(mapping.values())
+                    if not match.isdisjoint(seen):  # skip intersected groups
+                        continue
+                    seen.update(match)
+
+                    for n, fix in af.items():
+                        n = mapping[n]
+                        for key, value in fix.items():
+                            getattr(self, key)[n] = value
+                    for n, m, b in bf:
+                        n = mapping[n]
+                        m = mapping[m]
+                        bonds[n][m]._Bond__order = b
+        if seen:
+            self.flush_cache()
+
+    def __prepare_rings(self: 'MoleculeContainer'):
         atoms = self._atoms
         charges = self._charges
         radicals = self._radicals
         bonds = self._bonds
+        hydrogens = self._hydrogens
 
-        rings = defaultdict(set)  # aromatic skeleton
+        rings = defaultdict(list)  # aromatic skeleton
         pyroles = set()
 
         double_bonded = set()
@@ -137,7 +275,7 @@ class Aromatize:
             for m, bond in m_bond.items():
                 bo = bond.order
                 if bo == 4:
-                    rings[n].add(m)
+                    rings[n].append(m)
                 elif bo == 2:
                     double_bonded.add(n)
                 elif bo == 3:
@@ -147,9 +285,39 @@ class Aromatize:
             return rings, pyroles, set()
         elif not triple_bonded.isdisjoint(rings):
             raise InvalidAromaticRing('triple bonds connected to rings')
-        elif any(len(ms) not in (2, 3) for ms in rings.values()):
+
+        copy_rings = {n: ms.copy() for n, ms in rings.items()}
+        for r in self.sssr:
+            if set(r).issubset(rings):
+                n, *_, m = r
+                if n not in rings[m]:  # fix invalid structures: c1ccc-cc1
+                    rings[m].append(n)
+                    rings[n].append(m)
+                elif m in copy_rings[n]:
+                    copy_rings[n].remove(m)
+                    copy_rings[m].remove(n)
+                for n, m in zip(r, r[1:]):
+                    if n not in rings[m]:
+                        rings[m].append(n)
+                        rings[n].append(m)
+                    elif m in copy_rings[n]:
+                        copy_rings[n].remove(m)
+                        copy_rings[m].remove(n)
+
+        if any(len(ms) not in (2, 3) for ms in rings.values()):
             raise InvalidAromaticRing('not in ring aromatic bond or hypercondensed rings')
 
+        # fix invalid smiles: c1ccccc1c2ccccc2 instead of c1ccccc1-c2ccccc2
+        seen = set()
+        for n, ms in copy_rings.items():
+            if ms:
+                seen.add(n)
+                for m in ms:
+                    if m not in seen:
+                        rings[n].remove(m)
+                        rings[m].remove(n)
+                        bonds[n][m]._Bond__order = 1
+
         double_bonded &= rings.keys()
         if any(len(rings[n]) != 2 for n in double_bonded):  # double bonded never condensed
             raise InvalidAromaticRing('quinone valence error')
@@ -190,7 +358,13 @@ class Aromatize:
                         else:  # P(III) or P(V)H
                             pyroles.add(n)
                     elif ab == 2:
-                        pyroles.add(n)
+                        ah = hydrogens[n]
+                        if not ah:
+                            pyroles.add(n)
+                        elif ah == 1:  # only pyrole
+                            double_bonded.add(n)
+                        else:
+                            raise InvalidAromaticRing
                     elif ab != 4 or an != 15:  # P(V) in ring
                         raise InvalidAromaticRing
                 elif ac == -1:  # pyrole only
@@ -207,10 +381,19 @@ class Aromatize:
                 elif ab != 3:  # not pyridine oxyde
                     raise InvalidAromaticRing
             elif an == 8:  # furan
-                if ab == 2 and (ac == 0 and not radicals[n] or ac == 1 and radicals[n]):
-                    double_bonded.add(n)
+                if ab == 2:
+                    if ac == 0:
+                        if radicals[n]:
+                            raise InvalidAromaticRing('radical oxygen')
+                        double_bonded.add(n)
+                    elif ac == 1:
+                        if radicals[n]:  # furan cation-radical
+                            double_bonded.add(n)
+                        # pyrylium
+                    else:
+                        raise InvalidAromaticRing('invalid oxygen charge')
                 else:
-                    raise InvalidAromaticRing
+                    raise InvalidAromaticRing('Triple-bonded oxygen')
             elif an in (16, 34, 52):  # thiophene
                 if n not in double_bonded:  # not sulphoxyde or sulphone
                     if ab == 2:
@@ -256,7 +439,7 @@ class Aromatize:
                 raise InvalidAromaticRing(f'only B, C, N, P, O, S, Se, Te possible, not: {atoms[n].atomic_symbol}')
         return rings, pyroles, double_bonded
 
-    def __kekule_patch(self, patch):
+    def __kekule_patch(self: 'MoleculeContainer', patch):
         bonds = self._bonds
         atoms = set()
         for n, m, b in patch:
@@ -268,6 +451,7 @@ class Aromatize:
             self._calc_implicit(n)
 
     def __kekule_full(self):
+        self.__fix_oxides()  # fix pyridine n-oxyde
         rings, pyroles, double_bonded = self.__prepare_rings()
         atoms = set(rings)
         components = []
@@ -425,5 +609,71 @@ class Aromatize:
         if nether_yielded:
             raise InvalidAromaticRing(f'kekule form not found for: {list(rings)}')
 
+    @cached_property
+    def __freaks(self):
+        rules = []
+
+        q = query.QueryContainer()
+        q.add_atom('N', neighbors=2)
+        q.add_atom('A')
+        q.add_atom('A')
+        q.add_atom('A')
+        q.add_atom('A')
+        q.add_bond(1, 2, 1)
+        q.add_bond(2, 3, 2)
+        q.add_bond(3, 4, 1)
+        q.add_bond(4, 5, 4)
+        q.add_bond(1, 5, 1)
+        rules.append(q)
+
+        q = query.QueryContainer()
+        q.add_atom('N', neighbors=2)
+        q.add_atom('A')
+        q.add_atom('A')
+        q.add_atom('A')
+        q.add_atom('A')
+        q.add_bond(1, 2, 1)
+        q.add_bond(2, 3, 4)
+        q.add_bond(3, 4, 1)
+        q.add_bond(4, 5, 4)
+        q.add_bond(1, 5, 1)
+        rules.append(q)
+
+        return rules
+
+    @cached_property
+    def __oxyde_rules(self):
+        rules = []
+
+        # Aromatic N-Oxide
+        #
+        #  : N :  >>  : [N+] :
+        #    \\           \
+        #     O           [O-]
+        #
+        q = query.QueryContainer()
+        q.add_atom('N', neighbors=3, hybridization=4)
+        q.add_atom('O', neighbors=1)
+        q.add_bond(1, 2, 2)
+        atom_fix = {1: {'_charges': 1}, 2: {'_charges': -1, '_hybridizations': 1}}
+        bonds_fix = ((1, 2, 1),)
+        rules.append((q, atom_fix, bonds_fix))
+
+        # Aromatic N-Nitride?
+        #
+        #  : N :  >>  : [N+] :
+        #    \\           \
+        #     N           [N-]
+        #
+        q = query.QueryContainer()
+        q.add_atom('N', neighbors=3, hybridization=4)
+        q.add_atom('N', neighbors=(1, 2), hybridization=2)
+        q.add_bond(1, 2, 2)
+        atom_fix = {1: {'_charges': 1}, 2: {'_charges': -1, '_hybridizations': 1}}
+        bonds_fix = ((1, 2, 1),)
+        rules.append((q, atom_fix, bonds_fix))
+
+        return rules
+
 
 __all__ = ['Aromatize']
