@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-#  Copyright 2017-2020 Ramil Nugmanov <nougmanoff@protonmail.com>
+#  Copyright 2017-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
 #  Copyright 2019 Timur Gimadiev <timur.gimadiev@gmail.com>
 #  This file is part of CGRtools.
 #
@@ -17,14 +17,16 @@
 #  You should have received a copy of the GNU Lesser General Public License
 #  along with this program; if not, see <https://www.gnu.org/licenses/>.
 #
-from CachedMethods import cached_method
+from CachedMethods import cached_method, cached_property
 from collections import defaultdict
 from hashlib import sha512
 from itertools import count, product
+from random import random
+from typing import Tuple
 
 
 charge_str = {-4: '-4', -3: '-3', -2: '-2', -1: '-', 0: '0', 1: '+', 2: '+2', 3: '+3', 4: '+4'}
-order_str = {1: '', 2: '=', 3: '#', 4: ':', 8: '~', None: '.'}
+order_str = {1: '-', 2: '=', 3: '#', 4: ':', 8: '~', None: '.'}
 organic_set = {'C', 'N', 'O', 'P', 'S', 'F', 'Cl', 'Br', 'I', 'B'}
 hybridization_str = {4: 'a', 3: 't', 2: 'd', 1: 's', None: 'n'}
 dyn_order_str = {(None, 1): '[.>-]', (None, 2): '[.>=]', (None, 3): '[.>#]', (None, 4): '[.>:]', (None, 8): '[.>~]',
@@ -46,7 +48,7 @@ class Smiles:
 
     @cached_method
     def __str__(self):
-        return ''.join(self._smiles(self.atoms_order.get))
+        return ''.join(self._smiles(self._smiles_order))
 
     def __format__(self, format_spec):
         """
@@ -57,8 +59,13 @@ class Smiles:
             !s - Disable stereo marks.
             !h - Disable hybridization marks in queries. Returns non-unique signature.
             !n - Disable neighbors marks in queries. Returns non-unique signature.
-            !r - Use aromatic bonds instead aromatic atoms.
+            !H - Disable hydrogens marks in queries. Returns non-unique signature.
+            !R - Disable rings marks in queries. Returns non-unique signature.
+            !t - Disable heteroatoms marks in queries. Returns non-unique signature.
+            A - Use aromatic bonds instead aromatic atoms.
             m - Set atom mapping.
+            r - Generate random-ordered smiles.
+            o - Old canonic ordering algorithm.
 
             Combining possible. Order independent. Another keys ignored.
         """
@@ -72,11 +79,27 @@ class Smiles:
                 kwargs['hybridization'] = False
             if '!n' in format_spec:
                 kwargs['neighbors'] = False
-            if '!r' in format_spec:
+            if 'A' in format_spec:
                 kwargs['aromatic'] = False
             if 'm' in format_spec:
                 kwargs['mapping'] = True
-            return ''.join(self._smiles(self.atoms_order.get, **kwargs))
+            if '!H' in format_spec:
+                kwargs['hydrogens'] = False
+            if '!R' in format_spec:
+                kwargs['rings'] = False
+            if '!t' in format_spec:
+                kwargs['heteroatoms'] = False
+            if 'r' in format_spec:
+                kwargs['random'] = True
+
+                def w(x):
+                    return random()
+            elif 'o' in format_spec:
+                kwargs['old_order'] = True
+                w = self.atoms_order.get
+            else:
+                w = self._smiles_order
+            return ''.join(self._smiles(w, **kwargs))
         return str(self)
 
     def __eq__(self, other):
@@ -90,37 +113,66 @@ class Smiles:
     def __bytes__(self):
         return sha512(str(self).encode()).digest()
 
+    @cached_property
+    def smiles_atoms_order(self) -> Tuple[int, ...]:
+        """
+        Atoms order in canonic SMILES.
+        """
+        smiles, order = self._smiles(self._smiles_order, _return_order=True)
+        self.__dict__['__cached_method___str__'] = ''.join(smiles)  # cache smiles also
+        return tuple(order)
+
     def _smiles(self, weights, *, asymmetric_closures=False, open_parenthesis='(', close_parenthesis=')',
-                delimiter='.', **kwargs):
+                delimiter='.', _return_order=False, **kwargs):
         if not self._atoms:
             return []
         bonds = self._bonds
         atoms_set = set(self._atoms)
+        seen = {}
         cycles = count()
         casted_cycles = {}
         string = []
+        order = []
         if asymmetric_closures:
             visited_bond = set()
 
-        while True:
-            groups = defaultdict(int)
-            for n in atoms_set:
-                groups[weights(n)] += 1
+        groups = defaultdict(int)
+        for n in atoms_set:
+            groups[weights(n)] -= 1
 
-            def mod_weights(x):
-                # precedence of:
+        if kwargs.get('random', False):
+            mod_weights_start = mod_weights = weights
+        elif kwargs.get('old_order', False):
+            def mod_weights_start(x):
                 lb = len(bonds[x])
                 if lb:
-                    return (groups[weights(x)],  # rare groups
+                    return (-groups[weights(x)],  # rare groups
                             -lb,  # more neighbors
                             lb / len({weights(x) for x in bonds[x]}),  # more unique neighbors
                             weights(x))  # smallest weight
                 else:
-                    return groups[weights(x)], weights(x)  # rare groups > smallest weight
+                    return -groups[weights(x)], weights(x)  # rare groups > smallest weight
 
-            start = min(atoms_set, key=mod_weights)
+            def mod_weights(x):
+                lb = len(bonds[x])
+                return (-groups[weights(x)],  # rare groups
+                        -lb,  # more neighbors
+                        lb / len({weights(x) for x in bonds[x]}),  # more unique neighbors
+                        weights(x),  # smallest weight
+                        seen[x])  # BFS nearest to starting
+        else:
+            def mod_weights_start(x):
+                return (groups[weights(x)],  # common groups
+                        weights(x))  # smallest weight
 
-            seen = {start: 0}
+            def mod_weights(x):
+                return (groups[weights(x)],  # common groups
+                        weights(x),  # smallest weight
+                        seen[x])  # BFS nearest to starting
+
+        while True:
+            start = min(atoms_set, key=mod_weights_start)
+            seen[start] = 0
             queue = [(start, 1)]
             while queue:
                 n, d = queue.pop(0)
@@ -128,14 +180,6 @@ class Smiles:
                     queue.append((m, d + 1))
                     seen[m] = d
 
-            def mod_weights(x):
-                lb = len(bonds[x])
-                return (groups[weights(x)],  # rare groups
-                        -lb,  # more neighbors
-                        lb / len({weights(x) for x in bonds[x]}),  # more unique neighbors
-                        weights(x),  # smallest weight
-                        seen[x])  # BFS nearest to starting
-
             # modified NX dfs with cycle detection
             stack = [(start, len(atoms_set), iter(sorted(bonds[start], key=mod_weights)))]
             visited = {start: []}  # predecessors for stereo. atom: (visited[atom], *edges[atom])
@@ -156,8 +200,9 @@ class Smiles:
                             front = bonds[child].keys() - {parent}
                             if front:
                                 stack.append((child, depth_now - 1, iter(sorted(front, key=mod_weights))))
-                    elif child not in disconnected:
-                        disconnected.add(parent)
+                    elif (child, parent) not in disconnected:
+                        disconnected.add((parent, child))
+                        disconnected.add((child, parent))
                         cycle = next(cycles)
                         tokens[parent].append((child, cycle))
                         tokens[child].append((parent, cycle))
@@ -209,6 +254,7 @@ class Smiles:
             for token in smiles:
                 if isinstance(token, int):  # atoms
                     string.append(self._format_atom(token, adjacency=visited, **kwargs))
+                    order.append(token)
                     if token in tokens:
                         for m, c in tokens[token]:
                             if asymmetric_closures:
@@ -230,6 +276,8 @@ class Smiles:
                 string.append(delimiter)
             else:
                 break
+        if _return_order:
+            return string, order
         return string
 
     @staticmethod
@@ -240,6 +288,10 @@ class Smiles:
 class MoleculeSmiles(Smiles):
     __slots__ = ()
 
+    @property
+    def _smiles_order(self):
+        return self._chiral_morgan.__getitem__
+
     def _format_atom(self, n, adjacency, **kwargs):
         atom = self._atoms[n]
         charge = self._charges[n]
@@ -257,7 +309,7 @@ class MoleculeSmiles(Smiles):
 
         if kwargs.get('stereo', True):
             if n in self._atoms_stereo:
-                if ih and next(x for x in adjacency) == n:
+                if ih and next(x for x in adjacency) == n:  # first atom in smiles has reversed chiral mark
                     smi[3] = '@@' if self._translate_tetrahedron_sign(n, adjacency[n]) else '@'
                 else:
                     smi[3] = '@' if self._translate_tetrahedron_sign(n, adjacency[n]) else '@@'
@@ -341,6 +393,10 @@ class MoleculeSmiles(Smiles):
 class CGRSmiles(Smiles):
     __slots__ = ()
 
+    @property
+    def _smiles_order(self):
+        return self.atoms_order.__getitem__
+
     def _format_atom(self, n, **kwargs):
         atom = self._atoms[n]
         charge = self._charges[n]
@@ -381,11 +437,18 @@ class CGRSmiles(Smiles):
 class QuerySmiles(Smiles):
     __slots__ = ()
 
+    @property
+    def _smiles_order(self):
+        return self.atoms_order.__getitem__
+
     def _format_atom(self, n, **kwargs):
         atom = self._atoms[n]
         charge = self._charges[n]
         hybridization = self._hybridizations[n]
         neighbors = self._neighbors[n]
+        hydrogens = self._hydrogens[n]
+        rings = self._rings_sizes[n]
+        heteroatoms = self._heteroatoms[n]
 
         if atom.isotope:
             smi = ['[', str(atom.isotope), atom.atomic_symbol]
@@ -395,32 +458,46 @@ class QuerySmiles(Smiles):
         if kwargs.get('stereo', True) and n in self._atoms_stereo:  # carbon only
             smi.append('@' if self._translate_tetrahedron_sign(n, kwargs['adjacency'][n]) else '@@')
 
-        if kwargs.get('hybridization', True) and hybridization:
-            smi.append(';')
-            smi.append(''.join(hybridization_str[x] for x in hybridization))
-            if kwargs.get('neighbors', True) and neighbors:
-                smi.append(''.join(str(x) for x in neighbors))
-            smi.append(';')
-        elif kwargs.get('neighbors', True) and neighbors:
-            smi.append(';')
+        if kwargs.get('neighbors', True) and neighbors:
+            smi.append(';D')
             smi.append(''.join(str(x) for x in neighbors))
-            smi.append(';')
+
+        if kwargs.get('hydrogens', True) and hydrogens:
+            smi.append(';H')
+            smi.append(''.join(str(x) for x in hydrogens))
+
+        if kwargs.get('rings', True) and rings:
+            smi.append(';r')
+            smi.append(''.join(str(x) for x in rings))
+
+        if kwargs.get('hybridization', True) and hybridization:
+            smi.append(';Z')
+            smi.append(''.join(hybridization_str[x] for x in hybridization))
+
+        if kwargs.get('heteroatoms', True) and heteroatoms:
+            smi.append(';W')
+            smi.append(''.join(str(x) for x in heteroatoms))
+
+        if self._radicals[n]:
+            smi.append(';*')
 
         if charge:
             smi.append(charge_str[charge])
-        if self._radicals[n]:
-            smi.append('*')
 
         smi.append(']')
         return ''.join(smi)
 
     def _format_bond(self, n, m, **kwargs):
-        return order_str[self._bonds[n][m].order]
+        return ','.join(order_str[x] for x in self._bonds[n][m].order)
 
 
 class QueryCGRSmiles(Smiles):
     __slots__ = ()
 
+    @property
+    def _smiles_order(self):
+        return self.atoms_order.__getitem__
+
     def _format_atom(self, n, **kwargs):
         atom = self._atoms[n]
         charge = self._charges[n]
