@@ -19,12 +19,12 @@
 #
 from collections import defaultdict
 from functools import reduce
-from itertools import permutations
+from itertools import permutations, chain
 from io import StringIO, TextIOWrapper
 from logging import warning
 from operator import or_
 from pathlib import Path
-from re import split, compile, fullmatch
+from re import split, compile, fullmatch, findall, search
 from traceback import format_exc
 from typing import Union, List, Dict
 from warnings import warn
@@ -48,6 +48,7 @@ from ..exceptions import IncorrectSmiles, IsChiral, NotChiral, ValenceError
 # 9: up down bond
 # 10: dynamic bond
 # 11: dynamic atom
+# 12: dynamic aromatic atom
 
 replace_dict = {'-': 1, '=': 2, '#': 3, ':': 4, '~': 8, '.': None, '(': 2, ')': 3}
 charge_dict = {'+': 1, '+1': 1, '++': 2, '+2': 2, '+3': 3, '+++': 3, '+4': 4, '++++': 4,
@@ -65,8 +66,11 @@ dyn_charge_dict.update(tmp)
 dyn_radical_dict = {'*': (True,), '*>^': (True, ('radical', None)), '^>*': (False, ('radical', None))}
 
 atom_re = compile(r'([1-9][0-9]{0,2})?([A-IK-PR-Zacnopsb][a-ik-pr-vy]?)(@@|@)?(H[1-4]?)?([+-][1-4+-]?)?(:[0-9]{1,4})?')
-dyn_atom_re = compile(r'([1-9][0-9]{0,2})?([A-IK-PR-Z][a-ik-pr-vy]?)([+-0][1-4+-]?(>[+-0][1-4+-]?)?)?([*^](>[*^])?)?')
+dyn_atom_re = compile(r'([1-9][0-9]{0,2})?([A-IK-PR-Zacnopsb][a-ik-pr-vy]?)([+-0][1-4+-]?(>[+-0][1-4+-]?)?)?'
+                      r'([*^](>[*^])?)?')
 delimiter = compile(r'[=:]')
+cx_fragments = compile(r'f:(?:[0-9]+(?:\.[0-9]+)+)(?:,(?:[0-9]+(?:\.[0-9]+)+))*')
+cx_radicals = compile(r'\^[1-7]:[0-9]+(?:,[0-9]+)*')
 
 
 class SMILESRead(CGRRead):
@@ -184,22 +188,38 @@ class SMILESRead(CGRRead):
         if not smi:
             self._info('empty smiles')
             return {}
-        elif data and data[0].startswith('|f:'):
-            try:
-                contract = [sorted(int(x) for x in x.split('.')) for x in data[0][3:-1].split(',')]
-            except ValueError:
-                self._info(f'invalid cxsmiles fragments description: {data[0]}')
-                contract = None
-            else:
+        elif data and data[0].startswith('|') and data[0].endswith('|'):
+            fr = search(cx_fragments, data[0])
+            if fr is not None:
+                contract = [sorted(int(x) for x in x.split('.')) for x in fr.group()[2:].split(',')]
                 if len({x for x in contract for x in x}) < len([x for x in contract for x in x]):
                     self._info(f'collisions in cxsmiles fragments description: {data[0]}')
                     contract = None
                 elif any(x[0] < 0 for x in contract):
                     self._info(f'invalid cxsmiles fragments description: {data[0]}')
                     contract = None
-                else:
+
+                radicals = [int(x) for x in findall(cx_radicals, data[0]) for x in x[3:].split(',')]
+                if any(x < 0 for x in radicals):
+                    self._info(f'invalid cxsmiles radicals description: {data[0]}')
+                    radicals = []
+                if len(set(radicals)) != len(radicals):
+                    self._info(f'collisions in cxsmiles radicals description: {data[0]}')
+                    radicals = []
+                data = data[1:]
+            else:
+                radicals = [int(x) for x in findall(cx_radicals, data[0]) for x in x[3:].split(',')]
+                if radicals:
+                    if any(x < 0 for x in radicals):
+                        self._info(f'invalid cxsmiles radicals description: {data[0]}')
+                        radicals = []
+                    if len(set(radicals)) != len(radicals):
+                        self._info(f'collisions in cxsmiles radicals description: {data[0]}')
+                        radicals = []
                     data = data[1:]
+                contract = None
         else:
+            radicals = []
             contract = None
 
         if self.__header is None:
@@ -256,6 +276,12 @@ class SMILESRead(CGRRead):
                 self._info(f'record consist errors:\n{format_exc()}')
                 return meta
 
+            if radicals:
+                atom_map = dict(enumerate(a for m in chain(record['reactants'], record['reagents'], record['products'])
+                                          for a in m['atoms']))
+                for x in radicals:
+                    atom_map[x]['is_radical'] = True
+
             try:
                 container = self._convert_reaction(record)
             except ValueError:
@@ -307,6 +333,8 @@ class SMILESRead(CGRRead):
                 return meta
 
             record['meta'].update(meta)
+            for x in radicals:
+                record['atoms'][x]['is_radical'] = True
             try:
                 container = self._convert_structure(record)
             except ValueError:
@@ -330,17 +358,26 @@ class SMILESRead(CGRRead):
             if hc is None:  # aromatic rings or valence errors. just store given H count.
                 hydrogens[n] = h
             elif hc != h:  # H count mismatch. try radical state of atom.
-                radicals[n] = True
-                calc_implicit(n)
-                if hydrogens[n] != h:  # radical state also has errors.
+                if radicals[n]:
                     if self._ignore:
-                        radicals[n] = False  # reset radical state
                         hydrogens[n] = h  # set parsed hydrogens count
                         self._info(f'implicit hydrogen count ({h}) mismatch with '
                                    f'calculated ({hc}) on atom {n}. calculated count replaced.')
                     else:
                         raise ValueError(f'implicit hydrogen count ({h}) mismatch with '
                                          f'calculated ({hc}) on atom {n}.')
+                else:
+                    radicals[n] = True
+                    calc_implicit(n)
+                    if hydrogens[n] != h:  # radical state also has errors.
+                        if self._ignore:
+                            radicals[n] = False  # reset radical state
+                            hydrogens[n] = h  # set parsed hydrogens count
+                            self._info(f'implicit hydrogen count ({h}) mismatch with '
+                                       f'calculated ({hc}) on atom {n}. calculated count replaced.')
+                        else:
+                            raise ValueError(f'implicit hydrogen count ({h}) mismatch with '
+                                             f'calculated ({hc}) on atom {n}.')
 
         if self.__ignore_stereo or not molecule['stereo_atoms'] and not molecule['stereo_bonds']:
             return mol
@@ -536,9 +573,9 @@ class SMILESRead(CGRRead):
                         except KeyError:
                             raise IncorrectSmiles(f'invalid dynamic bond token {{{token}}}')
                     else:  # dynamic atom token
-                        out.append((11, cls.__dynatom_parse(token)))
+                        out.append(cls.__dynatom_parse(token))
                 elif '*' in token:  # CGR atom radical mark
-                    out.append((11, cls.__dynatom_parse(token)))
+                    out.append(cls.__dynatom_parse(token))
                 else:  # atom token
                     out.append(cls.__atom_parse(token))
             else:  # as is types: 1, 2, 3, 4, 6, 9
@@ -621,8 +658,13 @@ class SMILESRead(CGRRead):
         else:
             is_radical = False
 
-        return {'element': element, 'charge': charge, 'isotope': isotope, 'is_radical': is_radical,
-                'mapping': 0, 'x': 0., 'y': 0., 'z': 0., 'cgr': cgr}
+        if element in ('c', 'n', 'o', 'p', 's', 'as', 'se', 'b'):
+            _type = 12
+            element = element.capitalize()
+        else:
+            _type = 11
+        return _type, {'element': element, 'charge': charge, 'isotope': isotope, 'is_radical': is_radical,
+                       'mapping': 0, 'x': 0., 'y': 0., 'z': 0., 'cgr': cgr}
 
     def __parse_tokens(self, smiles):
         tokens = self._raw_tokenize(smiles)
@@ -633,9 +675,9 @@ class SMILESRead(CGRRead):
         strong_cycle = not self._ignore
         t1 = tokens[0][0]
         if t1 == 2:
-            if tokens[1][0] not in (0, 8, 11):
+            if tokens[1][0] not in (0, 8, 11, 12):
                 raise IncorrectSmiles('not atom started')
-        elif t1 not in (0, 8, 11):
+        elif t1 not in (0, 8, 11, 12):
             raise IncorrectSmiles('not atom started')
 
         atoms = []
@@ -721,7 +763,7 @@ class SMILESRead(CGRRead):
                             raise IncorrectSmiles('not equal cycle bonds')
                     else:
                         bt = 1
-                        b = 4 if atoms_types[last_num] == atoms_types[a] == 8 else 1
+                        b = 4 if atoms_types[last_num] in (8, 12) and atoms_types[a] in (8, 12) else 1
 
                     if bt == 1:
                         bonds.append((last_num, a, b))
@@ -736,7 +778,7 @@ class SMILESRead(CGRRead):
                 if atoms:
                     if not previous:
                         bt = 1
-                        b = 4 if atoms_types[last_num] == token_type == 8 else 1
+                        b = 4 if token_type in (8, 12) and atoms_types[last_num] in (8, 12) else 1
                         order[last_num].append(atom_num)
                         order[atom_num].append(last_num)
                     else:
@@ -754,7 +796,7 @@ class SMILESRead(CGRRead):
                         bonds.append((atom_num, last_num, 8))
                         cgr.append(((atom_num, last_num), 'bond', b))
 
-                if token_type == 11:
+                if token_type in (11, 12):
                     cgr.extend((atom_num, *x) for x in token.pop('cgr'))
                 else:
                     stereo = token.pop('stereo')
@@ -781,7 +823,7 @@ class SMILESRead(CGRRead):
         stereo_bonds = {n: ms for n, ms in stereo_bonds.items() if len(ms) == 1 or len(ms) == set(ms.values())}
         mol = {'atoms': atoms, 'bonds': bonds, 'order': order,
                'stereo_bonds': stereo_bonds, 'stereo_atoms': stereo_atoms, 'hydrogens': hydrogens, 'meta': {}}
-        if cgr or any(x == 11 for x in atoms_types):
+        if cgr or any(x in (11, 12) for x in atoms_types):
             mol['cgr'] = cgr
         return mol
 
