@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-#  Copyright 2019, 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
+#  Copyright 2019-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
 #  This file is part of CGRtools.
 #
 #  CGRtools is free software; you can redistribute it and/or modify
@@ -18,9 +18,10 @@
 #
 from importlib.util import find_spec
 from .functional_groups import functional_groups
+from .grid import grid_depict
 
 
-__all__ = ['functional_groups']
+__all__ = ['functional_groups', 'grid_depict']
 
 
 if find_spec('rdkit'):
